/*
 * CylinderLib.c++
 *
 *     Cylinder Library functions
 *
 *
 * Copyright 1996, 1997, Silicon Graphics, Inc.
 * ALL RIGHTS RESERVED
 *
 * UNPUBLISHED -- Rights reserved under the copyright laws of the United
 * States.   Use of a copyright notice is precautionary only and does not
 * imply publication or disclosure.
 *
 * U.S. GOVERNMENT RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in FAR 52.227.19(c)(2) or subparagraph (c)(1)(ii) of the Rights
 * in Technical Data and Computer Software clause at DFARS 252.227-7013 and/or
 * in similar or successor clauses in the FAR, or the DOD or NASA FAR
 * Supplement.  Contractor/manufacturer is Silicon Graphics, Inc.,
 * 2011 N. Shoreline Blvd. Mountain View, CA 94039-7311.
 *
 * THE CONTENT OF THIS WORK CONTAINS CONFIDENTIAL AND PROPRIETARY
 * INFORMATION OF SILICON GRAPHICS, INC. ANY DUPLICATION, MODIFICATION,
 * DISTRIBUTION, OR DISCLOSURE IN ANY FORM, IN WHOLE, OR IN PART, IS STRICTLY
 * PROHIBITED WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SILICON
 * GRAPHICS, INC.
 */
/**************************************************************************\
 *
 * OpenMOIV - C++ library for molecular visualization using Inventor.
 * Copyright (C) 2001-2003 Universitat Pompeu Fabra - Barcelona (Spain)
 * 
 * Developers: Interactive Technology Group (GTI)
 * Team: Josep Blat, Eduard Gonzalez, Sergi Gonzalez,
 *       Daniel Soto, Alejandro Ramirez, Oscar Civit.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details (see the file 
 * LICENSE.LGPL at the root directory).
 *
 * REMARK: This library is a derived product.
 *         You need also to accept all other applicable licenses.
 * 
 * Homepage: http://www.tecn.upf.es/openMOIV/
 * Contact:  openmoiv@upf.es
 *
\**************************************************************************/

#ident "$Revision: 1.3 $"

#include <assert.h>

#include <math.h>

/*
	Porting to Windows:
	Jamie Doornbos	2001/07/15
	Daniel Soto	2002/02/05
*/
#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <Inventor/SbColor.h>
#include <Inventor/SbLinear.h>
#include "CylinderLib.h"

// Returns S or T texture coord for point on top or bottom of
// cylinder, given x or z coord  (this is for flat caps)
#define BOT_TEX_S(x)    ((x) * .5 + .5)
#define BOT_TEX_T(z)    ((z) * .5 + .5)
#define TOP_TEX_S(x)    BOT_TEX_S(x)
#define TOP_TEX_T(z)    (1.0 - BOT_TEX_T(z))

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Generate the unit cylinder
//

void
cylinderLibUnitCylinder(SbBool doCaps, int numSections, int numSides,
	int &cylinderNormCount, int &cylinderVertCount,
	SbVec2f *ringCoords,
	SbVec3f *&cylinderNormals, SbVec3f *&cylinderVertices)
//
////////////////////////////////////////////////////////////////////////
{
    int         size;
    int         side, section;
    float       yTop, yBot, dy;
    float       outerRadius, innerRadius, dRadius;
    SbVec3f     pt, norm;

    if (cylinderNormals != NULL) {
        delete [] cylinderNormals;
    }
    // Sides
    size = numSections*numSides + numSections;
    // Top and Bottom
    size += 2;
    cylinderNormals = new SbVec3f[size];

    if (cylinderVertices != NULL) {
        delete [] cylinderVertices;
    }
    // Sides
    size = 2*numSections*(numSides+1);
    // Caps
    size += 2 * (2*(numSections-1) * (numSides + 1) + 
            2 + numSides);
    cylinderVertices = new SbVec3f[size];

    cylinderNormCount = cylinderVertCount = 0;

    // Draw each section of sides as a triangle mesh, from top to bottom
    yTop = 1.0;
    dy   = -2.0 / numSections;

    for (section = 0; section < numSections; section++) {
        yBot = yTop + dy;
        for (side = 0; side < numSides; side++) {
            pt[0] = ringCoords[side][0];
            pt[2] = ringCoords[side][1];

            // Deal with normal
            norm.setValue(pt[0], 0.0, pt[2]);
            cylinderNormals[cylinderNormCount++] = norm;

            // Point at bottom of section
            pt[1] = yBot;
            cylinderVertices[cylinderVertCount++] = pt;

            // Point at top of section
            pt[1] = yTop;
            cylinderVertices[cylinderVertCount++] = pt;
        }

        // Join end of strip back to beginning
        side = 0;
        pt[0] = ringCoords[side][0];
        pt[2] = ringCoords[side][1];

        // Deal with normal
        norm.setValue(pt[0], 0.0, pt[2]);
        cylinderNormals[cylinderNormCount++] = norm;

        // Point at bottom of section
        pt[1] = yBot;
        cylinderVertices[cylinderVertCount++] = pt;

        // Point at top of section
        pt[1] = yTop;
        cylinderVertices[cylinderVertCount++] = pt;

        // Prepare for next section down
        yTop = yBot;
    }

    // Draw top face as a series of concentric rings. The number of
    // rings is the same as the number of sections of the sides of the
    // cylinder.

    if (doCaps) {
        // TOP
        norm.setValue(0.0, 1.0, 0.0);
        pt[1] = 1.0;

        cylinderNormals[cylinderNormCount++] = norm;

        // Start at the outside and work in
        outerRadius = 1.0;
        dRadius     = -1.0 / numSections;

        for (section = numSections - 1; section > 0; --section) {

            innerRadius = outerRadius + dRadius;

            // Innermost ring will be a triangle fan calculated later.
            // Other rings are triangle strips.

            for (side = 0; side < numSides; side++) {
                // Send points on outer and inner rings
                pt[0] = outerRadius * ringCoords[side][0];
                pt[2] = outerRadius * ringCoords[side][1];
                cylinderVertices[cylinderVertCount++] = pt;
                pt[0] = innerRadius * ringCoords[side][0];
                pt[2] = innerRadius * ringCoords[side][1];
                cylinderVertices[cylinderVertCount++] = pt;
            }

            // Join end of strip back to beginning
            pt[0] = outerRadius * ringCoords[0][0];
            pt[2] = outerRadius * ringCoords[0][1];
            cylinderVertices[cylinderVertCount++] = pt;
            pt[0] = innerRadius * ringCoords[0][0];
            pt[2] = innerRadius * ringCoords[0][1];
            cylinderVertices[cylinderVertCount++] = pt;

            // Prepare for next ring
            outerRadius = innerRadius;
        }

        // Top innermost ring
        // Innermost ring is treated as a triangle fan. This not
        // only gets better shading (because the center vertex is
        // sent), but also avoids the problem of having a polygon
        // with too many vertices.

        outerRadius = -dRadius;

        // Center point comes first
        pt[0] = pt[2] = 0.0;
        cylinderVertices[cylinderVertCount++] = pt;

        // Send all vertices around ring. Go in reverse order
        // so that vertex ordering is correct
        for (side = numSides - 1; side >= 0; side--) {
            pt[0] = outerRadius * ringCoords[side][0];
            pt[2] = outerRadius * ringCoords[side][1];
            cylinderVertices[cylinderVertCount++] = pt;
        }
        // Send first vertex again
        pt[0] = outerRadius * ringCoords[numSides - 1][0];
        pt[2] = outerRadius * ringCoords[numSides - 1][1];
        cylinderVertices[cylinderVertCount++] = pt;

        // BOTTOM
        // Draw bottom face the same way as the top
        norm.setValue(0.0, -1.0, 0.0);
        pt[1] = -1.0;

        cylinderNormals[cylinderNormCount++] = norm;

        // Start at the outside and work in
        outerRadius = 1.0;
        dRadius     = -1.0 / numSections;
        for (section = numSections - 1; section > 0; --section) {

            innerRadius = outerRadius + dRadius;

            // Innermost ring will be a triangle fan calculated later.
            // Other rings are triangle strips.

            // Go in reverse order so that vertex ordering is correct
            for (side = numSides - 1; side >= 0; side--) {
                // Send points on outer and inner rings
                pt[0] = outerRadius * ringCoords[side][0];
                pt[2] = outerRadius * ringCoords[side][1];
                cylinderVertices[cylinderVertCount++] = pt;
                pt[0] = innerRadius * ringCoords[side][0];
                pt[2] = innerRadius * ringCoords[side][1];
                cylinderVertices[cylinderVertCount++] = pt;
            }

            // Join end of strip back to beginning
            side = numSides - 1;
            pt[0] = outerRadius * ringCoords[side][0];
            pt[2] = outerRadius * ringCoords[side][1];
            cylinderVertices[cylinderVertCount++] = pt;
            pt[0] = innerRadius * ringCoords[side][0];
            pt[2] = innerRadius * ringCoords[side][1];
            cylinderVertices[cylinderVertCount++] = pt;

            // Prepare for next ring
            outerRadius = innerRadius;
        } // end for (section = numSections - 1 ...

        // Bottom innermost ring.
        // Innermost ring is drawn as a triangle fan. This not
        // only gets better shading (because the center vertex is
        // sent), but also avoids the problem of having a polygon
        // with too many vertices.

        outerRadius = -dRadius;

        // Center point comes first
        pt[1] = -1.0;
        pt[0] = pt[2] = 0.0;
        cylinderVertices[cylinderVertCount++] = pt;

        // Send all vertices around ring
        for (side = 0; side < numSides; side++) {
            pt[0] = outerRadius * ringCoords[side][0];
            pt[2] = outerRadius * ringCoords[side][1];
            cylinderVertices[cylinderVertCount++] = pt;
        }
        // Send first vertex again
        pt[0] = outerRadius * ringCoords[0][0];
        pt[2] = outerRadius * ringCoords[0][1];
        cylinderVertices[cylinderVertCount++] = pt;

    } // end if (doCaps)
}

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Generate the unit cylinder caps for rounded cylinders
//

void
cylinderLibUnitCylinderCaps(int numSections, int numSides,
	int &cylinderCapNormCount, int &cylinderCapVertCount,
	int &cylinderBottomStartNorm, int &cylinderBottomStartVert,
	SbVec2f *ringCoords,
	SbVec3f *&cylinderCapNormals, SbVec3f *&cylinderCapVertices)
//
////////////////////////////////////////////////////////////////////////
{
#ifdef DEBUG
printf ("In cylinderLibUnitCylinderCaps\n");
#endif
    int     section;
    SbVec3f norm, pt;
    float   innerRadius, outerRadius;
    float   dY[3], dRadius[3];
    int     side;

    if (cylinderCapNormals != NULL) delete [] cylinderCapNormals;
    if (cylinderCapVertices != NULL) delete [] cylinderCapVertices;

    int size = 2*(numSections-1)*numSides + 2*(numSections-1);
    size += numSides + 2;
    cylinderCapNormals = new SbVec3f[2*size];
    cylinderCapVertices = new SbVec3f[2*size];

    // Start at the outside and work in
    outerRadius = 1.0;
    dY[2]           = 0.0;
    dY[1]           = sin(M_PI / 6.0);
    dY[0]           = sin(M_PI / 3.0);
    dRadius[2] = 1.0;
    dRadius[1] = cos(M_PI / 6.0);
    dRadius[0] = cos(M_PI / 3.0);

    cylinderCapNormCount = cylinderCapVertCount = 0;

    for (section = numSections - 1; section > 0; --section) {

        outerRadius = dRadius[section];
        innerRadius = dRadius[section-1];

        // Innermost ring will be a triangle fan calculated later.
        // Other rings are triangle strips.

        for (side = 0; side < numSides; side++) {
            // Send points on outer and inner rings
            pt[0] = outerRadius * ringCoords[side][0];
            pt[1] = dY[section];
            pt[2] = outerRadius * ringCoords[side][1];
            norm.setValue(pt[0], pt[1], pt[2]);
            norm.normalize();
            cylinderCapNormals[cylinderCapNormCount++] = norm;
            cylinderCapVertices[cylinderCapVertCount++] = pt;

            pt[0] = innerRadius * ringCoords[side][0];
            pt[1] = dY[section-1];
            pt[2] = innerRadius * ringCoords[side][1];
            norm.setValue(pt[0], pt[1], pt[2]);
            norm.normalize();
            cylinderCapNormals[cylinderCapNormCount++] = norm;
            cylinderCapVertices[cylinderCapVertCount++] = pt;
        }

        // Join end of strip back to beginning
        pt[0] = outerRadius * ringCoords[0][0];
        pt[1] = dY[section];
        pt[2] = outerRadius * ringCoords[0][1];
        norm.setValue(pt[0], pt[1], pt[2]);
        norm.normalize();
        cylinderCapNormals[cylinderCapNormCount++] = norm;
        cylinderCapVertices[cylinderCapVertCount++] = pt;

        pt[0] = innerRadius * ringCoords[0][0];
        pt[1] = dY[section-1];;
        pt[2] = innerRadius * ringCoords[0][1];
        norm.setValue(pt[0], pt[1], pt[2]);
        norm.normalize();
        cylinderCapNormals[cylinderCapNormCount++] = norm;
        cylinderCapVertices[cylinderCapVertCount++] = pt;
    }

    // Cap innermost ring
    // Innermost ring is treated as a triangle fan. This not
    // only gets better shading (because the center vertex is
    // sent), but also avoids the problem of having a polygon
    // with too many vertices.

    // Center point comes first
    pt[0] = pt[2] = 0.0;
    pt[1] = 1.0;
    cylinderCapNormals[cylinderCapNormCount++] = pt;
    cylinderCapVertices[cylinderCapVertCount++] = pt;

    // Send all vertices around ring. Go in reverse order
    // so that vertex ordering is correct
    outerRadius = dRadius[0];
    pt[1] = dY[0];
    for (side = numSides - 1; side >= 0; side--) {
        pt[0] = outerRadius * ringCoords[side][0];
        pt[2] = outerRadius * ringCoords[side][1];
        norm.setValue(pt[0], pt[1], pt[2]);
        norm.normalize();
        cylinderCapNormals[cylinderCapNormCount++] = norm;
        cylinderCapVertices[cylinderCapVertCount++] = pt;
    }
    // Send first vertex again
    pt[0] = outerRadius * ringCoords[numSides - 1][0];
    pt[2] = outerRadius * ringCoords[numSides - 1][1];
    norm.setValue(pt[0], pt[1], pt[2]);
    norm.normalize();
    cylinderCapNormals[cylinderCapNormCount++] = norm;
    cylinderCapVertices[cylinderCapVertCount++] = pt;

    // BOTTOM
    // Draw bottom face the same way as the top
    // Start at the outside and work in

    cylinderBottomStartNorm = cylinderCapNormCount;
    cylinderBottomStartVert = cylinderCapVertCount;
    for (section = numSections - 1; section > 0; --section) {

        outerRadius = dRadius[section];
        innerRadius = dRadius[section-1];

        // Innermost ring will be a triangle fan calculated later.
        // Other rings are triangle strips.

        // Go in reverse order so that vertex ordering is correct
        for (side = numSides - 1; side >= 0; side--) {
            // Send points on outer and inner rings
            pt[0] = outerRadius * ringCoords[side][0];
            pt[1] = -dY[section];
            pt[2] = outerRadius * ringCoords[side][1];
            norm.setValue(pt[0], pt[1], pt[2]);
            norm.normalize();
            cylinderCapNormals[cylinderCapNormCount++] = norm;
            cylinderCapVertices[cylinderCapVertCount++] = pt;

            pt[0] = innerRadius * ringCoords[side][0];
            pt[1] = -dY[section-1];
            pt[2] = innerRadius * ringCoords[side][1];
            norm.setValue(pt[0], pt[1], pt[2]);
            norm.normalize();
            cylinderCapNormals[cylinderCapNormCount++] = norm;
            cylinderCapVertices[cylinderCapVertCount++] = pt;
        }

        // Join end of strip back to beginning
        side = numSides - 1;
        pt[0] = outerRadius * ringCoords[side][0];
        pt[1] = -dY[section];
        pt[2] = outerRadius * ringCoords[side][1];
        norm.setValue(pt[0], pt[1], pt[2]);
        norm.normalize();
        cylinderCapNormals[cylinderCapNormCount++] = norm;
        cylinderCapVertices[cylinderCapVertCount++] = pt;

        pt[0] = innerRadius * ringCoords[side][0];
        pt[1] = -dY[section-1];;
        pt[2] = innerRadius * ringCoords[side][1];
        norm.setValue(pt[0], pt[1], pt[2]);
        norm.normalize();
        cylinderCapNormals[cylinderCapNormCount++] = norm;
        cylinderCapVertices[cylinderCapVertCount++] = pt;

    } // end for (section = numSections - 1 ...

    // Bottom innermost ring.
    // Innermost ring is drawn as a triangle fan. This not
    // only gets better shading (because the center vertex is
    // sent), but also avoids the problem of having a polygon
    // with too many vertices.

    // Center point comes first
    pt[0] = pt[2] = 0.0;
    pt[1] = -1.0;
    cylinderCapNormals[cylinderCapNormCount++] = pt;
    cylinderCapVertices[cylinderCapVertCount++] = pt;

    // Send all vertices around ring
    outerRadius = dRadius[0];
    pt[1] = -dY[0];
    for (side = 0; side < numSides; side++) {
        pt[0] = outerRadius * ringCoords[side][0];
        pt[2] = outerRadius * ringCoords[side][1];
        norm.setValue(pt[0], pt[1], pt[2]);
        norm.normalize();
        cylinderCapNormals[cylinderCapNormCount++] = norm;
        cylinderCapVertices[cylinderCapVertCount++] = pt;
    }
    // Send first vertex again
    pt[0] = outerRadius * ringCoords[0][0];
    pt[2] = outerRadius * ringCoords[0][1];
    norm.setValue(pt[0], pt[1], pt[2]);
    norm.normalize();
    cylinderCapNormals[cylinderCapNormCount++] = norm;
    cylinderCapVertices[cylinderCapVertCount++] = pt;
}

#define DISP_NORMAL                                                           \
theNormMatrix.multDirMatrix(cylinderNormals[normCount++], vec);               \
glNormal3fv(vec.getValue());

#define DISP_VERTEX                                                           \
theMatrix.multVecMatrix(cylinderVertices[vertCount++], vec);                  \
glVertex3fv(vec.getValue());

#define DISP_FLATCAP_TOP_TEXTURE                                              \
vec = cylinderVertices[vertCount++];                                          \
s = TOP_TEX_S(vec[0]);                                                        \
t = TOP_TEX_T(vec[2]);                                                        \
glTexCoord2f(s, t);                                                           \
theMatrix.multVecMatrix(vec, tmpVec);                                         \
glVertex3fv(tmpVec.getValue());

#define DISP_FLATCAP_BOT_TEXTURE                                              \
vec = cylinderVertices[vertCount++];                                          \
s = BOT_TEX_S(vec[0]);                                                        \
t = BOT_TEX_T(vec[2]);                                                        \
glTexCoord2f(s, t);                                                           \
theMatrix.multVecMatrix(vec, tmpVec);                                         \
glVertex3fv(tmpVec.getValue());

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Issues GL calls to display a cylinder.
//

void
cylinderLibGenerateCylinderDisplayList(const SbMatrix &theMatrix,
    const SbBool doCaps, const SbBool doNormals, const SbBool doTextures, 
	const int numSections, const int numSides,
	const SbVec3f *cylinderNormals, const SbVec3f *cylinderVertices)
//
////////////////////////////////////////////////////////////////////////
{
    int     normCount, vertCount;
    int     section, side;
    float   s, t;
    float   ds, dt;
    float   tTop, tBot;
    SbVec3f vec;
    SbVec3f tmpVec;

    SbMatrix theNormMatrix = theMatrix.inverse().transpose();

    // Don't do normals or textures
    if (!doNormals && !doTextures) {
        // Do the cylinder sides
        normCount = 0;
        vertCount = 0;
        for (section = 0; section < numSections; section++) {
            glBegin(GL_TRIANGLE_STRIP);
            for (side = 0; side < numSides; side++) {
                DISP_VERTEX;
                DISP_VERTEX;
            }
            DISP_VERTEX;
            DISP_VERTEX;
            glEnd();
        }

        if (doCaps) {
            // TOP
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = 0; side < numSides; side++) {
                    DISP_VERTEX;
                    DISP_VERTEX;
                }
                DISP_VERTEX;
                DISP_VERTEX;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            DISP_VERTEX;
            for (side = numSides - 1; side >= 0; side--) {
                DISP_VERTEX;
            }
            DISP_VERTEX;
            glEnd();

            
            // BOTTOM
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = numSides - 1; side >= 0; side--) {
                    DISP_VERTEX;
                    DISP_VERTEX;
                }
                DISP_VERTEX;
                DISP_VERTEX;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            DISP_VERTEX;
            for (side = 0; side < numSides; side++) {
                DISP_VERTEX;
            }
            DISP_VERTEX;
            glEnd();
        }
    }
    else if (doNormals && !doTextures) {
        // Do the cylinder sides
        normCount = 0;
        vertCount = 0;
        for (section = 0; section < numSections; section++) {
            glBegin(GL_TRIANGLE_STRIP);
            for (side = 0; side < numSides; side++) {
                DISP_NORMAL;
                DISP_VERTEX;
                DISP_VERTEX;
            }
            DISP_NORMAL;
            DISP_VERTEX;
            DISP_VERTEX;
            glEnd();
        }

        if (doCaps) {
            // TOP
            DISP_NORMAL;
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = 0; side < numSides; side++) {
                    DISP_VERTEX;
                    DISP_VERTEX;
                }
                DISP_VERTEX;
                DISP_VERTEX;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            DISP_VERTEX;
            for (side = numSides - 1; side >= 0; side--) {
                DISP_VERTEX;
            }
            DISP_VERTEX;
            glEnd();
            
            // BOTTOM
            DISP_NORMAL;
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = numSides - 1; side >= 0; side--) {
                    DISP_VERTEX;
                    DISP_VERTEX;
                }
                DISP_VERTEX;
                DISP_VERTEX;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            DISP_VERTEX;
            for (side = 0; side < numSides; side++) {
                DISP_VERTEX;
            }
            DISP_VERTEX;
            glEnd();
        }
    }
    else if (!doNormals && doTextures) {
        // Do the cylinder sides
        normCount = 0;
        vertCount = 0;
        tTop = 1.0;
        dt = -1.0 / numSections;
        ds = -1.0 / numSides;
        for (section = 0; section < numSections; section++) {
            s = 1.0;
            tBot = tTop + dt;
            glBegin(GL_TRIANGLE_STRIP);
            for (side = 0; side < numSides; side++) {
                glTexCoord2f(s, tBot);
                DISP_VERTEX;

                glTexCoord2f(s, tTop);
                DISP_VERTEX;

                s += ds;
            }
            s = 0.0;
            glTexCoord2f(s, tBot);
            DISP_VERTEX;

            glTexCoord2f(s, tTop);
            DISP_VERTEX;
            glEnd();

            tTop = tBot;
        }

        if (doCaps) {
            // TOP
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = 0; side < numSides; side++) {
                    DISP_FLATCAP_TOP_TEXTURE;
                    DISP_FLATCAP_TOP_TEXTURE;
                }
                DISP_FLATCAP_TOP_TEXTURE;
                DISP_FLATCAP_TOP_TEXTURE;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            glTexCoord2f(0.5, 0.5);
            DISP_VERTEX;
            for (side = numSides - 1; side >= 0; side--) {
                DISP_FLATCAP_TOP_TEXTURE;
            }
            DISP_FLATCAP_TOP_TEXTURE;
            glEnd();

            
            // BOTTOM
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = numSides - 1; side >= 0; side--) {
                    DISP_FLATCAP_BOT_TEXTURE;
                    DISP_FLATCAP_BOT_TEXTURE;
                }
                DISP_FLATCAP_BOT_TEXTURE;
                DISP_FLATCAP_BOT_TEXTURE;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            glTexCoord2f(0.5, 0.5);
            DISP_VERTEX;
            for (side = 0; side < numSides; side++) {
                DISP_FLATCAP_BOT_TEXTURE;
            }
            DISP_FLATCAP_BOT_TEXTURE;
            glEnd();
        }
    }
    else if (doNormals && doTextures) {
        // Do the cylinder sides
        normCount = 0;
        vertCount = 0;
        tTop = 1.0;
        dt = -1.0 / numSections;
        ds = -1.0 / numSides;
        for (section = 0; section < numSections; section++) {
            s = 1.0;
            tBot = tTop + dt;
            glBegin(GL_TRIANGLE_STRIP);
            for (side = 0; side < numSides; side++) {
                DISP_NORMAL;
                glTexCoord2f(s, tBot);
                DISP_VERTEX;

                glTexCoord2f(s, tTop);
                DISP_VERTEX;

                s += ds;
            }
            s = 0.0;
            DISP_NORMAL;
            glTexCoord2f(s, tBot);
            DISP_VERTEX;

            glTexCoord2f(s, tTop);
            DISP_VERTEX;
            glEnd();

            tTop = tBot;
        }

        if (doCaps) {
            // TOP
            DISP_NORMAL;
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = 0; side < numSides; side++) {
                    DISP_FLATCAP_TOP_TEXTURE;
                    DISP_FLATCAP_TOP_TEXTURE;
                }
                DISP_FLATCAP_TOP_TEXTURE;
                DISP_FLATCAP_TOP_TEXTURE;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            glTexCoord2f(0.5, 0.5);
            DISP_VERTEX;
            for (side = numSides - 1; side >= 0; side--) {
                DISP_FLATCAP_TOP_TEXTURE;
            }
            DISP_FLATCAP_TOP_TEXTURE;
            glEnd();

            
            // BOTTOM
            DISP_NORMAL;
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = numSides - 1; side >= 0; side--) {
                    DISP_FLATCAP_BOT_TEXTURE;
                    DISP_FLATCAP_BOT_TEXTURE;
                }
                DISP_FLATCAP_BOT_TEXTURE;
                DISP_FLATCAP_BOT_TEXTURE;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            glTexCoord2f(0.5, 0.5);
            DISP_VERTEX;
            for (side = 0; side < numSides; side++) {
                DISP_FLATCAP_BOT_TEXTURE;
            }
            DISP_FLATCAP_BOT_TEXTURE;
            glEnd();
        }
    }
}

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Issues GL calls to display a cylinder with colors interpolated
//    between the ends.
//

void
cylinderLibGenerateColoredCylinderDisplayList(const SbMatrix &theMatrix,
    const SbBool doCaps, const SbBool doNormals, const SbBool doTextures, 
    const int numSections, const int numSides, 
    const SbColor color1, const SbColor color2,
	const SbVec3f *cylinderNormals, const SbVec3f *cylinderVertices)
//
////////////////////////////////////////////////////////////////////////
{
    int     normCount, vertCount;
    int     section, side;
    float   alpha;
    float   s, t;
    float   ds, dt;
    float   tTop, tBot;
    float   tmpFloat;
    SbVec3f vec;
    SbVec3f tmpVec;

    const float *lColor1;
    const float *lColor2;

    SbMatrix theNormMatrix = theMatrix.inverse().transpose();

    alpha = 1.0 / (float)numSections;
    // Don't do normals or textures
    if (!doNormals && !doTextures) {
        // Do the cylinder sides
        normCount = 0;
        vertCount = 0;
        for (section = 0; section < numSections; section++) {
            tmpFloat = (numSections - section - 1)*alpha;
            lColor1 = ((1-tmpFloat)*color1 + (tmpFloat*color2)).getValue();
            tmpFloat += alpha;
            lColor2 = ((1-tmpFloat)*color1 + (tmpFloat*color2)).getValue();
            glBegin(GL_TRIANGLE_STRIP);
            for (side = 0; side < numSides; side++) {
                glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor1);
                DISP_VERTEX;
                glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor2);
                DISP_VERTEX;
            }
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor1);
            DISP_VERTEX;
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor2);
            DISP_VERTEX;
            glEnd();
        }

        if (doCaps) {
            lColor1 = color1.getValue();
            lColor2 = color2.getValue();
            // TOP
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor2);
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = 0; side < numSides; side++) {
                    DISP_VERTEX;
                    DISP_VERTEX;
                }
                DISP_VERTEX;
                DISP_VERTEX;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            DISP_VERTEX;
            for (side = numSides - 1; side >= 0; side--) {
                DISP_VERTEX;
            }
            DISP_VERTEX;
            glEnd();

            
            // BOTTOM
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor1);
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = numSides - 1; side >= 0; side--) {
                    DISP_VERTEX;
                    DISP_VERTEX;
                }
                DISP_VERTEX;
                DISP_VERTEX;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            DISP_VERTEX;
            for (side = 0; side < numSides; side++) {
                DISP_VERTEX;
            }
            DISP_VERTEX;
            glEnd();
        }
    }
    else if (doNormals && !doTextures) {
        // Do the cylinder sides
        normCount = 0;
        vertCount = 0;
        for (section = 0; section < numSections; section++) {
            tmpFloat = (numSections - section - 1)*alpha;
            lColor1 = ((1-tmpFloat)*color1 + (tmpFloat*color2)).getValue();
            tmpFloat += alpha;
            lColor2 = ((1-tmpFloat)*color1 + (tmpFloat*color2)).getValue();
            glBegin(GL_TRIANGLE_STRIP);
            for (side = 0; side < numSides; side++) {
                DISP_NORMAL;
                glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor1);
                DISP_VERTEX;
                glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor2);
                DISP_VERTEX;
            }
            DISP_NORMAL;
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor1);
            DISP_VERTEX;
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor2);
            DISP_VERTEX;
            glEnd();
        }

        if (doCaps) {
            lColor1 = color1.getValue();
            lColor2 = color2.getValue();
            // TOP
            DISP_NORMAL;
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor2);
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = 0; side < numSides; side++) {
                    DISP_VERTEX;
                    DISP_VERTEX;
                }
                DISP_VERTEX;
                DISP_VERTEX;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            DISP_VERTEX;
            for (side = numSides - 1; side >= 0; side--) {
                DISP_VERTEX;
            }
            DISP_VERTEX;
            glEnd();
            
            // BOTTOM
            DISP_NORMAL;
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor1);
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = numSides - 1; side >= 0; side--) {
                    DISP_VERTEX;
                    DISP_VERTEX;
                }
                DISP_VERTEX;
                DISP_VERTEX;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            DISP_VERTEX;
            for (side = 0; side < numSides; side++) {
                DISP_VERTEX;
            }
            DISP_VERTEX;
            glEnd();
        }
    }
    else if (!doNormals && doTextures) {
        // Do the cylinder sides
        normCount = 0;
        vertCount = 0;
        tTop = 1.0;
        dt = -1.0 / numSections;
        ds = -1.0 / numSides;
        for (section = 0; section < numSections; section++) {
            tmpFloat = (numSections - section - 1)*alpha;
            lColor1 = ((1-tmpFloat)*color1 + (tmpFloat*color2)).getValue();
            tmpFloat += alpha;
            lColor2 = ((1-tmpFloat)*color1 + (tmpFloat*color2)).getValue();

            s = 1.0;
            tBot = tTop + dt;
            glBegin(GL_TRIANGLE_STRIP);
            for (side = 0; side < numSides; side++) {
                glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor1);
                glTexCoord2f(s, tBot);
                DISP_VERTEX;

                glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor2);
                glTexCoord2f(s, tTop);
                DISP_VERTEX;

                s += ds;
            }
            s = 0.0;
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor1);
            glTexCoord2f(s, tBot);
            DISP_VERTEX;

            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor2);
            glTexCoord2f(s, tTop);
            DISP_VERTEX;
            glEnd();

            tTop = tBot;
        }

        if (doCaps) {
            lColor1 = color1.getValue();
            lColor2 = color2.getValue();
            // TOP
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor2);
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = 0; side < numSides; side++) {
                    DISP_FLATCAP_TOP_TEXTURE;
                    DISP_FLATCAP_TOP_TEXTURE;
                }
                DISP_FLATCAP_TOP_TEXTURE;
                DISP_FLATCAP_TOP_TEXTURE;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            glTexCoord2f(0.5, 0.5);
            DISP_VERTEX;
            for (side = numSides - 1; side >= 0; side--) {
                DISP_FLATCAP_TOP_TEXTURE;
            }
            DISP_FLATCAP_TOP_TEXTURE;
            glEnd();

            
            // BOTTOM
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor1);
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = numSides - 1; side >= 0; side--) {
                    DISP_FLATCAP_BOT_TEXTURE;
                    DISP_FLATCAP_BOT_TEXTURE;
                }
                DISP_FLATCAP_BOT_TEXTURE;
                DISP_FLATCAP_BOT_TEXTURE;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            glTexCoord2f(0.5, 0.5);
            DISP_VERTEX;
            for (side = 0; side < numSides; side++) {
                DISP_FLATCAP_BOT_TEXTURE;
            }
            DISP_FLATCAP_BOT_TEXTURE;
            glEnd();
        }
    }
    else if (doNormals && doTextures) {
        // Do the cylinder sides
        normCount = 0;
        vertCount = 0;
        tTop = 1.0;
        dt = -1.0 / numSections;
        ds = -1.0 / numSides;
        for (section = 0; section < numSections; section++) {
            tmpFloat = (numSections - section - 1)*alpha;
            lColor1 = ((1-tmpFloat)*color1 + (tmpFloat*color2)).getValue();
            tmpFloat += alpha;
            lColor2 = ((1-tmpFloat)*color1 + (tmpFloat*color2)).getValue();

            s = 1.0;
            tBot = tTop + dt;
            glBegin(GL_TRIANGLE_STRIP);
            for (side = 0; side < numSides; side++) {
                DISP_NORMAL;
                glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor1);
                glTexCoord2f(s, tBot);
                DISP_VERTEX;

                glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor2);
                glTexCoord2f(s, tTop);
                DISP_VERTEX;

                s += ds;
            }
            s = 0.0;
            DISP_NORMAL;
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor1);
            glTexCoord2f(s, tBot);
            DISP_VERTEX;

            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor2);
            glTexCoord2f(s, tTop);
            DISP_VERTEX;
            glEnd();

            tTop = tBot;
        }

        if (doCaps) {
            lColor1 = color1.getValue();
            lColor2 = color2.getValue();
            // TOP
            DISP_NORMAL;
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor2);
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = 0; side < numSides; side++) {
                    DISP_FLATCAP_TOP_TEXTURE;
                    DISP_FLATCAP_TOP_TEXTURE;
                }
                DISP_FLATCAP_TOP_TEXTURE;
                DISP_FLATCAP_TOP_TEXTURE;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            glTexCoord2f(0.5, 0.5);
            DISP_VERTEX;
            for (side = numSides - 1; side >= 0; side--) {
                DISP_FLATCAP_TOP_TEXTURE;
            }
            DISP_FLATCAP_TOP_TEXTURE;
            glEnd();

            
            // BOTTOM
            DISP_NORMAL;
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lColor1);
            for (section = numSections - 1; section > 0; --section) {
                glBegin(GL_TRIANGLE_STRIP);
                for (side = numSides - 1; side >= 0; side--) {
                    DISP_FLATCAP_BOT_TEXTURE;
                    DISP_FLATCAP_BOT_TEXTURE;
                }
                DISP_FLATCAP_BOT_TEXTURE;
                DISP_FLATCAP_BOT_TEXTURE;
                glEnd();
            }
            glBegin(GL_TRIANGLE_FAN);
            glTexCoord2f(0.5, 0.5);
            DISP_VERTEX;
            for (side = 0; side < numSides; side++) {
                DISP_FLATCAP_BOT_TEXTURE;
            }
            DISP_FLATCAP_BOT_TEXTURE;
            glEnd();
        }
    }
}

#undef DISP_NORMAL
#undef DISP_VERTEX
#undef DISP_FLATCAP_TOP_TEXTURE
#undef DISP_FLATCAP_BOT_TEXTURE

#define DISP_ROUNDCAP_NORMAL                                                  \
theNormMatrix.multDirMatrix(cylinderCapNormals[normCount++], vec);            \
glNormal3fv(vec.getValue());

#define DISP_ROUNDCAP_VERTEX                                                  \
theMatrix.multVecMatrix(cylinderCapVertices[vertCount++], vec);               \
glVertex3fv(vec.getValue());

#define DISP_ROUNDCAP_TOP_TEXTURE                                             \
vec = cylinderCapVertices[vertCount++];                                       \
s = TOP_TEX_S(vec[0]);                                                        \
t = TOP_TEX_T(vec[2]);                                                        \
glTexCoord2f(s, t);                                                           \
theMatrix.multVecMatrix(vec, tmpVec);                                         \
glVertex3fv(tmpVec.getValue());

#define DISP_ROUNDCAP_BOT_TEXTURE                                             \
vec = cylinderCapVertices[vertCount++];                                       \
s = BOT_TEX_S(vec[0]);                                                        \
t = BOT_TEX_T(vec[2]);                                                        \
glTexCoord2f(s, t);                                                           \
theMatrix.multVecMatrix(vec, tmpVec);                                         \
glVertex3fv(tmpVec.getValue());

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Issues GL calls to display rounded cylinder caps.
//

void
cylinderLibGenerateCylinderCapDisplayList(const SbMatrix &theMatrix,
    const SbBool bottom, const SbBool doNormals, const SbBool doTextures,
	const int numSections, const int numSides,
	const int cylinderBottomStartNorm, const int cylinderBottomStartVert,
	const SbVec3f *cylinderCapNormals, const SbVec3f *cylinderCapVertices)
//
////////////////////////////////////////////////////////////////////////
{
    int     side;
	int	    section;
    int     normCount;
    int     vertCount;
    float   s, t;
    SbVec3f vec;
    SbVec3f tmpVec;

    SbMatrix theNormMatrix = theMatrix.inverse().transpose();

    if (bottom) {
        normCount = cylinderBottomStartNorm;
        vertCount = cylinderBottomStartVert;
    }
    else {
        normCount = vertCount = 0;
    }

    // Don't do normals or textures
    if (!doNormals && !doTextures) {
	    for (section = numSections-1; section > 0; --section) {
	        glBegin(GL_TRIANGLE_STRIP);
	        for (int side = 0; side < numSides; side++) {
                DISP_ROUNDCAP_VERTEX;
                DISP_ROUNDCAP_VERTEX;
	        }
            DISP_ROUNDCAP_VERTEX;
            DISP_ROUNDCAP_VERTEX;
	        glEnd();
	    }
	    glBegin(GL_TRIANGLE_FAN);
        DISP_ROUNDCAP_VERTEX;
	    for (side = numSides - 1; side >= 0; side--) {
            DISP_ROUNDCAP_VERTEX;
	    }
        DISP_ROUNDCAP_VERTEX;
	    glEnd();
	}
    else if (doNormals && !doTextures) {
	    for (section = numSections-1; section > 0; --section) {
	        glBegin(GL_TRIANGLE_STRIP);
	        for (int side = 0; side < numSides; side++) {
                DISP_ROUNDCAP_NORMAL;
                DISP_ROUNDCAP_VERTEX;

                DISP_ROUNDCAP_NORMAL;
                DISP_ROUNDCAP_VERTEX;
	        }
            DISP_ROUNDCAP_NORMAL;
            DISP_ROUNDCAP_VERTEX;

            DISP_ROUNDCAP_NORMAL;
            DISP_ROUNDCAP_VERTEX;
	        glEnd();
	    }
	    glBegin(GL_TRIANGLE_FAN);
        DISP_ROUNDCAP_NORMAL;
        DISP_ROUNDCAP_VERTEX;
	    for (side = numSides - 1; side >= 0; side--) {
            DISP_ROUNDCAP_NORMAL;
            DISP_ROUNDCAP_VERTEX;
	    }
        DISP_ROUNDCAP_NORMAL;
        DISP_ROUNDCAP_VERTEX;
	    glEnd();
	}
    else if (!doNormals && doTextures) {
        if (bottom) {
    	    for (section = numSections-1; section > 0; --section) {
    	        glBegin(GL_TRIANGLE_STRIP);
    	        for (int side = 0; side < numSides; side++) {
                    DISP_ROUNDCAP_BOT_TEXTURE;
                    DISP_ROUNDCAP_BOT_TEXTURE;
    	        }
                DISP_ROUNDCAP_BOT_TEXTURE;
                DISP_ROUNDCAP_BOT_TEXTURE;
    	        glEnd();
    	    }
    	    glBegin(GL_TRIANGLE_FAN);
            glTexCoord2f(0.5, 0.5);
            DISP_ROUNDCAP_VERTEX;
    	    for (side = numSides - 1; side >= 0; side--) {
                DISP_ROUNDCAP_BOT_TEXTURE;
    	    }
            DISP_ROUNDCAP_BOT_TEXTURE;
    	    glEnd();
        }
        else {
    	    for (section = numSections-1; section > 0; --section) {
    	        glBegin(GL_TRIANGLE_STRIP);
    	        for (int side = 0; side < numSides; side++) {
                    DISP_ROUNDCAP_TOP_TEXTURE;
                    DISP_ROUNDCAP_TOP_TEXTURE;
    	        }
                DISP_ROUNDCAP_TOP_TEXTURE;
                DISP_ROUNDCAP_TOP_TEXTURE;
    	        glEnd();
    	    }
    	    glBegin(GL_TRIANGLE_FAN);
            glTexCoord2f(0.5, 0.5);
            DISP_ROUNDCAP_VERTEX;
    	    for (side = numSides - 1; side >= 0; side--) {
                DISP_ROUNDCAP_TOP_TEXTURE;
    	    }
            DISP_ROUNDCAP_TOP_TEXTURE;
    	    glEnd();
        }
	}
    else if (doNormals && doTextures) {
        if (bottom) {
    	    for (section = numSections-1; section > 0; --section) {
    	        glBegin(GL_TRIANGLE_STRIP);
    	        for (int side = 0; side < numSides; side++) {
                    DISP_ROUNDCAP_NORMAL;
                    DISP_ROUNDCAP_BOT_TEXTURE;
    
                    DISP_ROUNDCAP_NORMAL;
                    DISP_ROUNDCAP_BOT_TEXTURE;
    	        }
                DISP_ROUNDCAP_NORMAL;
                DISP_ROUNDCAP_BOT_TEXTURE;
    
                DISP_ROUNDCAP_NORMAL;
                DISP_ROUNDCAP_BOT_TEXTURE;
    	        glEnd();
    	    }
    	    glBegin(GL_TRIANGLE_FAN);
            DISP_ROUNDCAP_NORMAL;
            glTexCoord2f(0.5, 0.5);
            DISP_ROUNDCAP_VERTEX;
    	    for (side = numSides - 1; side >= 0; side--) {
                DISP_ROUNDCAP_NORMAL;
                DISP_ROUNDCAP_BOT_TEXTURE;
    	    }
            DISP_ROUNDCAP_NORMAL;
            DISP_ROUNDCAP_BOT_TEXTURE;
    	    glEnd();
        }
        else {
    	    for (section = numSections-1; section > 0; --section) {
    	        glBegin(GL_TRIANGLE_STRIP);
    	        for (int side = 0; side < numSides; side++) {
                    DISP_ROUNDCAP_NORMAL;
                    DISP_ROUNDCAP_TOP_TEXTURE;
    
                    DISP_ROUNDCAP_NORMAL;
                    DISP_ROUNDCAP_TOP_TEXTURE;
    	        }
                DISP_ROUNDCAP_NORMAL;
                DISP_ROUNDCAP_TOP_TEXTURE;
    
                DISP_ROUNDCAP_NORMAL;
                DISP_ROUNDCAP_TOP_TEXTURE;
    	        glEnd();
    	    }
    	    glBegin(GL_TRIANGLE_FAN);
            DISP_ROUNDCAP_NORMAL;
            glTexCoord2f(0.5, 0.5);
            DISP_ROUNDCAP_VERTEX;
    	    for (side = numSides - 1; side >= 0; side--) {
                DISP_ROUNDCAP_NORMAL;
                DISP_ROUNDCAP_TOP_TEXTURE;
    	    }
            DISP_ROUNDCAP_NORMAL;
            DISP_ROUNDCAP_TOP_TEXTURE;
    	    glEnd();
        }
	}
}

#undef DISP_ROUNDCAP_NORMAL
#undef DISP_ROUNDCAP_VERTEX
#undef DISP_ROUNDCAP_TOP_TEXTURE
#undef DISP_ROUNDCAP_BOT_TEXTURE


////////////////////////////////////////////////////////////////////////
//
// Description:
//    Computes number of sides per circular cross-section and number
//    of sections, based on complexity, then computes ring of x,z
//    coordinates around cylinder and stores in ringCoords.
//

void
cylinderLibComputeDepth(float complexity, 
	int &numSections, int &numSides)
//
////////////////////////////////////////////////////////////////////////
{
    // If complexity is between 0 and .5 (inclusive), use 1 section
    // and between 3 and 16 sides:
    if (complexity <= 0.5) {
        numSections = 1;
        numSides    = (int) (complexity * 26.0 + 3.0);
    }

    // If complexity is between .5 and 1, use between 1 and 8 sections
    // and between 16 and 64 sides:
    else {
        numSections = (int) (14.0 * complexity - 6.0);
        numSides    = (int) (complexity * 96.0 - 32.0);
    }
}

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Computes number of sides per circular cross-section and number
//    of sections, based on complexity, then computes ring of x,z
//    coordinates around cylinder and stores in ringCoords.
//

void
cylinderLibComputeRingCoords(int numSides, SbVec2f *&ringCoords)
//
////////////////////////////////////////////////////////////////////////
{
    float   theta, dTheta;
    int     side;

    if (ringCoords != NULL) delete [] ringCoords;
    ringCoords = new SbVec2f[numSides];

    // Compute x and z coordinates around ring
    theta  = 0.0;
    dTheta = 2.0 * M_PI / (numSides);
    for (side = 0; side < numSides; side++) {
        ringCoords[side].setValue(sin(theta), -cos(theta));
        theta += dTheta;
    }
}

#undef BOT_TEX_S
#undef BOT_TEX_T
#undef TOP_TEX_S
#undef TOP_TEX_T

