/*
 * SphereLib.c++
 *
 *     Sphere Library functions
 *
 *
 * Copyright 1996, 1997, Silicon Graphics, Inc.
 * ALL RIGHTS RESERVED
 *
 * UNPUBLISHED -- Rights reserved under the copyright laws of the United
 * States.   Use of a copyright notice is precautionary only and does not
 * imply publication or disclosure.
 *
 * U.S. GOVERNMENT RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in FAR 52.227.19(c)(2) or subparagraph (c)(1)(ii) of the Rights
 * in Technical Data and Computer Software clause at DFARS 252.227-7013 and/or
 * in similar or successor clauses in the FAR, or the DOD or NASA FAR
 * Supplement.  Contractor/manufacturer is Silicon Graphics, Inc.,
 * 2011 N. Shoreline Blvd. Mountain View, CA 94039-7311.
 *
 * THE CONTENT OF THIS WORK CONTAINS CONFIDENTIAL AND PROPRIETARY
 * INFORMATION OF SILICON GRAPHICS, INC. ANY DUPLICATION, MODIFICATION,
 * DISTRIBUTION, OR DISCLOSURE IN ANY FORM, IN WHOLE, OR IN PART, IS STRICTLY
 * PROHIBITED WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SILICON
 * GRAPHICS, INC.
 */
/**************************************************************************\
 *
 * OpenMOIV - C++ library for molecular visualization using Inventor.
 * Copyright (C) 2001-2003 Universitat Pompeu Fabra - Barcelona (Spain)
 * 
 * Developers: Interactive Technology Group (GTI)
 * Team: Josep Blat, Eduard Gonzalez, Sergi Gonzalez,
 *       Daniel Soto, Alejandro Ramirez, Oscar Civit.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details (see the file 
 * LICENSE.LGPL at the root directory).
 *
 * REMARK: This library is a derived product.
 *         You need also to accept all other applicable licenses.
 * 
 * Homepage: http://www.tecn.upf.es/openMOIV/
 * Contact:  openmoiv@upf.es
 *
\**************************************************************************/

#ident "$Revision: 1.3 $"

#include <assert.h>
#include <math.h>

/*
	Porting to Windows:
	Jamie Doornbos	2001/07/15
	Daniel Soto	2002/02/05
*/
#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <Inventor/SbLinear.h>
#include "SphereLib.h"

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Computes sphere tesselation depth based on current complexity. 
//

int
sphereLibComputeDepth(float complexity)
//
////////////////////////////////////////////////////////////////////////
{
    int     depth;

    // In object space, compute tesselation depth based only on
    // complexity value. We want the following correspondences between
    // complexity and depth:  
    //      0.0 -> 1
    //      0.5 -> 5
    //      1.0 -> 30
    // So we'll use linear ramps from 0 to .5 and from .5 to 1

    if (complexity < 0.5)
        depth = 1   + (int) ((complexity) * 9.0);
    else
        depth = -20 + (int) ((complexity) * 50.0);

    return depth;
}

static float tridata[6][3] = {
    {1, 0, 0},
    {-1, 0, 0},
    {0, 1, 0},
    {0, -1, 0},
    {0, 0, 1},
    {0, 0, -1}
};

static int triindex[8][3] = {
    {4, 3, 0}, {4, 1, 3},
    {4, 2, 1}, {4, 0, 2},

    {5, 0, 2}, {5, 2, 1},
    {5, 1, 3}, {5, 3, 0}
};

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Generates the vertices describing a unit sphere given a specified 
//    depth.  It allocates space for the vertices and also returns the
//    number of vertices generated in sphereVertCount.
//

void
sphereLibUnitSphere ( int depth, int &sphereVertCount, 
	SbVec3f *&sphereVertices )
//
////////////////////////////////////////////////////////////////////////
{
    int     i, j, k, t;
    int     octSize;
    float   *v0, *v1, *v2;
    float   w[3][3];
    SbVec3f vec;

    // Create vertices for a unit sphere.
    // For the unit sphere, there will be 8*depth*depth + 12*depth - 2
    // vertices that make up the main triangle strip.

    if (sphereVertices != NULL) delete [] sphereVertices;
    octSize = 8*depth*depth + 12*depth - 2;
    sphereVertices = new SbVec3f[octSize];

	sphereVertCount = 0;
    for (i = 0; i < depth; i++) { // each level
        for (t = 0; t <= 3; t++) { 
            // across each orig triangle in sphere
            // set triangle verticies 
            v0 = tridata[triindex[t][0]];
            v1 = tridata[triindex[t][1]];
            v2 = tridata[triindex[t][2]];

            for (j = 0; i + j < depth; j++) { // do one piece of row
                k = depth - i - j;
                w[0][0] = (i*v0[0]     + j*v1[0] +     k*v2[0])/depth;
                w[0][1] = (i*v0[1]     + j*v1[1] +     k*v2[1])/depth;
                w[0][2] = (i*v0[2]     + j*v1[2] +     k*v2[2])/depth;
                w[1][0] = ((i+1)*v0[0] + j*v1[0] + (k-1)*v2[0])/depth;
                w[1][1] = ((i+1)*v0[1] + j*v1[1] + (k-1)*v2[1])/depth;
                w[1][2] = ((i+1)*v0[2] + j*v1[2] + (k-1)*v2[2])/depth;
                w[2][0] = (i*v0[0] + (j+1)*v1[0] + (k-1)*v2[0])/depth;
                w[2][1] = (i*v0[1] + (j+1)*v1[1] + (k-1)*v2[1])/depth;
                w[2][2] = (i*v0[2] + (j+1)*v1[2] + (k-1)*v2[2])/depth;

                if (j == 0) {  // first ttri
                    if ((t == 0))  {
                        // very first ttri of row - output 1st,2nd vert
                        vec.setValue(w[0][0], w[0][1], w[0][2]);
                        vec.normalize();
                        sphereVertices[sphereVertCount++] = vec;

                        vec.setValue(w[1][0], w[1][1], w[1][2]);
                        vec.normalize();
                        sphereVertices[sphereVertCount++] = vec;
                    }
                    else  {
                        // first ttri in a sub-row 
                        vec = sphereVertices[sphereVertCount-2];
                        sphereVertices[sphereVertCount++] = vec;
                    }
                }

                // output lower left vert of ttri - 3rd vert
                vec.setValue(w[2][0], w[2][1], w[2][2]);
                vec.normalize();
                sphereVertices[sphereVertCount++] = vec;

                // upside-down ttri 
                // - never one at the end of a sub-row 
                //   or at the top of the tri
                //
                if ((i < (depth - 1)) && ((i+j) < (depth - 1))) {
                    w[0][0] = ((i+1)*v0[0] + (j+1)*v1[0] + (k-2)*v2[0])/depth;
                    w[0][1] = ((i+1)*v0[1] + (j+1)*v1[1] + (k-2)*v2[1])/depth;
                    w[0][2] = ((i+1)*v0[2] + (j+1)*v1[2] + (k-2)*v2[2])/depth;

                    vec.setValue(w[0][0], w[0][1], w[0][2]);
                    vec.normalize();
                    sphereVertices[sphereVertCount++] = vec;
                }
            } // end for (j = 0 ...
        } // end for (t = 0 ...

        // if not on last row do a swap
        if (i < (depth - 1)) {
            vec = sphereVertices[sphereVertCount-2];
            sphereVertices[sphereVertCount++] = vec;
        }
    } // end for (i = 0 ...

    // Generate the back of the sphere
    sphereLibHemiSymmetry(sphereVertCount, sphereVertices);
}

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Given the z+ hemisphere, mirror the vertices to z- for the other 
//    hemisphere
//

void
sphereLibHemiSymmetry(int &sphereVertCount, SbVec3f *&sphereVertices)
//
////////////////////////////////////////////////////////////////////////
{
    int i;
    int iend;
    float x, y, z;

    iend = sphereVertCount;
    for (i = 0; i < iend; i++) {
        sphereVertices[i].getValue(x, y, z);
        sphereVertices[sphereVertCount++].setValue(x, -y, -z);
    }
}

// Computes S and T texture coordinates from point on surface
#define COMPUTE_S_T(point, s, t)                                              \
    s = atan2f(point[0], point[2]) * .159 + .5;                               \
    t = atan2f(point[1],                                                      \
           sqrtf(point[0]*point[0] + point[2]*point[2])) * .318 + .5

;   
// Adjusts S texture coordinate in unstable areas
#define ADJUST_S(s, vec)                                                      \
    if (s < .001 && (vec[0] < 0.0 && vec[2] > 0.0))                           \
        s = 1.0;                                                              \
    else if (s > .999 && (vec[0] < 0.0 && vec[2] < 0.0))                      \
        s = 0.0

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Generate the sphere display list containing either 1 triangle strip
//    if hemispheres are used or 2 triangle strips if the full sphere
//    is used.
//

void
sphereLibGenerateSphereDisplayList(SbBool hemi, int sphereVertCount,
	SbVec3f *sphereVertices, SbBool doNormals, SbBool doTextures)
//
////////////////////////////////////////////////////////////////////////
{
#ifdef DEBUG
printf ("In sphereLibGenerateSphereDisplayList\n");
#endif
    int     i, vend;
    float   s,t;

    vend = sphereVertCount/2;

    // Don't do normals or textures
    if (!doNormals && !doTextures) {
        glBegin(GL_TRIANGLE_STRIP);
        for (i = 0; i < vend; i++) {
            glVertex3fv(sphereVertices[i].getValue());
        }
        glEnd();

        if (hemi == FALSE) {
            glBegin(GL_TRIANGLE_STRIP);
            for (i = vend; i < sphereVertCount; i++) {
                glVertex3fv(sphereVertices[i].getValue());
            }
            glEnd();
        }

    // Do normals but not textures.  This is the most common case.
    }
    else if (doNormals && !doTextures) {
        glBegin(GL_TRIANGLE_STRIP);
        for (i = 0; i < vend; i++) {
            glNormal3fv(sphereVertices[i].getValue());
            glVertex3fv(sphereVertices[i].getValue());
        }
        glEnd();

        if (hemi == FALSE) {
            glBegin(GL_TRIANGLE_STRIP);
            for (i = vend; i < sphereVertCount; i++) {
                glNormal3fv(sphereVertices[i].getValue());
                glVertex3fv(sphereVertices[i].getValue());
            }
            glEnd();
        }

    // Don't do normals but do textures
    }
    else if (!doNormals && doTextures) {
        glBegin(GL_TRIANGLE_STRIP);
        for (i = 0; i < vend; i++) {
            COMPUTE_S_T(sphereVertices[i].getValue(), s, t);
            ADJUST_S(s, sphereVertices[i].getValue());
            glTexCoord2f(s, t);
            glNormal3fv(sphereVertices[i].getValue());
            glVertex3fv(sphereVertices[i].getValue());
        }
        glEnd();

        if (hemi == FALSE) {
            glBegin(GL_TRIANGLE_STRIP);
            for (i = vend; i < sphereVertCount; i++) {
                COMPUTE_S_T(sphereVertices[i].getValue(), s, t);
                ADJUST_S(s, sphereVertices[i].getValue());
                glTexCoord2f(s, t);
                glNormal3fv(sphereVertices[i].getValue());
                glVertex3fv(sphereVertices[i].getValue());
            }
            glEnd();
        }

    // Do both normals and textures
    }
    else if (doNormals && doTextures) {
        glBegin(GL_TRIANGLE_STRIP);
        for (i = 0; i < vend; i++) {
            COMPUTE_S_T(sphereVertices[i].getValue(), s, t);
            ADJUST_S(s, sphereVertices[i].getValue());
            glTexCoord2f(s, t);
            glNormal3fv(sphereVertices[i].getValue());
            glVertex3fv(sphereVertices[i].getValue());
        }
        glEnd();

        if (hemi == FALSE) {
            glBegin(GL_TRIANGLE_STRIP);
            for (i = vend; i < sphereVertCount; i++) {
                COMPUTE_S_T(sphereVertices[i].getValue(), s, t);
                ADJUST_S(s, sphereVertices[i].getValue());
                glTexCoord2f(s, t);
                glNormal3fv(sphereVertices[i].getValue());
                glVertex3fv(sphereVertices[i].getValue());
            }
            glEnd();
        }
    }
}

#undef COMPUTE_S_T
#undef ADJUST_S
